﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing.Drawing2D;

namespace Ch5PenExample
{
    public partial class MainFormCh5Pen : Form
    {
        public MainFormCh5Pen()
        {
            InitializeComponent();
            SetStyle(ControlStyles.ResizeRedraw, true);
        }

        private void MainFormCh5Pen_Load(object sender, EventArgs e)
        {

        }

        enum DrawState { None, Compound, Join, Brushes, DrawShapes, FillShapes, 
            DrawCurves, DynamicBezier, ClosedBezier };
        DrawState drawState = DrawState.None;

        private void MainFormCh5Pen_Paint(object sender, PaintEventArgs e)
        {

        }

        private void compoundToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Text = "Compound Pen";
            drawState = DrawState.Compound;
            panelDraw.Invalidate();
        }

        private void DrawCompound(Graphics g)
        {
            using (Pen pen = new Pen(Color.Black, 20))
            {
                pen.CompoundArray = new float[] {
                            0.2f, 0.25f, 0.45f, 0.55f, 0.75f, .9f 
                        };
                Rectangle rect = panelDraw.ClientRectangle;
                rect.Inflate(-20, -20);
                g.DrawRectangle(pen, rect);
            }
        }

        private void FillShapes(Graphics graphics)
        {
            using (Pen pen = new Pen(Color.Black, 4))
            {
                int width = 120;
                int height = 60;
                int margin = 15;
                Rectangle rect = new Rectangle(margin, margin, width - 2 * margin, height - margin);
                graphics.DrawArc(pen, rect, 0, -180);
                graphics.TranslateTransform(width, 0);
                graphics.DrawBezier(pen, new Point(margin, height - margin),
                    new Point(width - margin, height - margin), new Point(margin, margin), new Point(width - margin, margin));
                graphics.TranslateTransform(width, 0);
                graphics.FillClosedCurve(Brushes.Gray, new Point[] {
                    new Point(width - margin, margin), new Point(margin, margin),
                    new Point(width - margin, height - margin), new Point(margin, height - margin)});
                graphics.DrawClosedCurve(pen, new Point[] {
                    new Point(width - margin, margin), new Point(margin, margin),
                    new Point(width - margin, height - margin), new Point(margin, height - margin)});
                graphics.TranslateTransform(-2 * width, height);
                graphics.DrawCurve(pen, new Point[] {
                    new Point(width - margin, margin), new Point(margin, margin),
                    new Point(width - margin, height - margin), new Point(margin, height - margin)});
                graphics.TranslateTransform(width, 0);
                graphics.FillEllipse(Brushes.Gray, rect);
                graphics.DrawEllipse(pen, rect);
                graphics.TranslateTransform(width, 0);
                graphics.DrawLines(pen, new Point[] {
                    new Point(width - margin, margin), new Point(margin, margin),
                    new Point(width - margin, height - margin), new Point(margin, height - margin)});
                graphics.TranslateTransform(-2 * width, height);
                graphics.FillPie(Brushes.Gray, rect, 0, -180);
                graphics.DrawPie(pen, rect, 0, -180);
                graphics.TranslateTransform(width, 0);
                graphics.FillPolygon(Brushes.Gray, new Point[] {
                    new Point(width - margin, margin), new Point(margin, margin),
                    new Point(width - margin, height - margin), new Point(margin, height - margin)});
                graphics.DrawPolygon(pen, new Point[] {
                    new Point(width - margin, margin), new Point(margin, margin),
                    new Point(width - margin, height - margin), new Point(margin, height - margin)});
                graphics.TranslateTransform(width, 0);
                graphics.FillRectangle(Brushes.Gray, rect);
                graphics.DrawRectangle(pen, rect);
            }
        }

        private void DrawShapes(Graphics graphics)
        {
            using (Pen pen = new Pen(Color.Black, 4))
            {
                int width = 120;
                int height = 60;
                int margin = 15;
                Rectangle rect = new Rectangle(margin, margin, width - 2*margin, height - margin);
                graphics.DrawArc(pen, rect, 0, -180);
                graphics.TranslateTransform(width, 0);
                graphics.DrawBezier(pen, new Point(margin, height - margin),
                    new Point(width - margin, height - margin), new Point(margin, margin), new Point(width - margin, margin));
                graphics.TranslateTransform(width, 0);
                graphics.DrawClosedCurve(pen, new Point[] {
                    new Point(width - margin, margin), new Point(margin, margin),
                    new Point(width - margin, height - margin), new Point(margin, height - margin)});
                graphics.TranslateTransform(-2*width, height);
                graphics.DrawCurve(pen, new Point[] {
                    new Point(width - margin, margin), new Point(margin, margin),
                    new Point(width - margin, height - margin), new Point(margin, height - margin)});
                graphics.TranslateTransform(width, 0);
                graphics.DrawEllipse(pen, rect);
                graphics.TranslateTransform(width, 0);
                graphics.DrawLines(pen, new Point[] {
                    new Point(width - margin, margin), new Point(margin, margin),
                    new Point(width - margin, height - margin), new Point(margin, height - margin)});
                graphics.TranslateTransform(-2*width, height);
                graphics.DrawPie(pen, rect, 0, -180);
                graphics.TranslateTransform(width, 0);
                graphics.DrawPolygon(pen, new Point[] {
                    new Point(width - margin, margin), new Point(margin, margin),
                    new Point(width - margin, height - margin), new Point(margin, height - margin)});
                graphics.TranslateTransform(width, 0);
                graphics.DrawRectangle(pen, rect);
            }
        }

        private void DrawBrushes(Graphics g)
        {
            using (LinearGradientBrush brush = new LinearGradientBrush(
                panelDraw.ClientRectangle,
                Color.Empty,
                Color.Empty,
                45f))
            {
                ColorBlend blend = new ColorBlend();
                blend.Colors = new Color[] { Color.Red, Color.Orange, Color.Yellow, 
                    Color.Blue, Color.Green, Color.Violet };
                blend.Positions = new float[] { 0f, 0.2f, 0.4f, 0.6f, 0.8f, 1f };
                brush.InterpolationColors = blend;
                using (Pen pen = new Pen(brush))
                {
                    DrawArtLines(g, pen);
                }
                StringFormat format = new StringFormat();
                format.LineAlignment = StringAlignment.Center;
                format.Alignment = StringAlignment.Center;
                g.DrawString("Fancy Pen", new Font(this.Font.FontFamily, 36), brush,
                    panelDraw.ClientRectangle, format);
            }
        }

        private void DrawArtLines(Graphics g, Pen pen)
        {
            int steps = 50;
            float delta = panelDraw.ClientRectangle.Width / (float)steps;
            PointF start = new Point(0, 0);
            PointF end = new Point(panelDraw.ClientRectangle.Width, 0);
            for (int i = 0; i < steps; i++)
            {
                start.Y += delta;
                g.DrawLine(pen, start, end);
                end.X -= delta;
            }
            start = new Point(0, 0);
            end = new Point(panelDraw.Width, 0);
            for (int i = 0; i < steps; i++)
            {
                end.Y += delta;
                g.DrawLine(pen, start, end);
                start.X += delta;
            }
            start = new Point(panelDraw.Width, 0);
            end = new Point(panelDraw.Width, panelDraw.Height);
            for (int i = 0; i < steps; i++)
            {
                end.X -= delta;
                g.DrawLine(pen, start, end);
                start.Y += delta;

            }
            start = new Point(panelDraw.Width, panelDraw.Height);
            end = new Point(0, panelDraw.Height);
            for (int i = 0; i < steps; i++)
            {
                end.Y -= delta;
                g.DrawLine(pen, start, end);
                start.X -= delta;
            }
        }

        private void DrawJoin(Graphics g)
        {
            int y = 20;
            using (Pen pen = new Pen(Color.Black, 10))
            {
                pen.CompoundArray = new float[] {
                    0f, .45f, .55f, 1f
                };
                foreach (LineJoin join in Enum.GetValues(typeof(LineJoin)))
                {
                    pen.LineJoin = join;
                    DrawJoinRect(g, pen, new Point(20, y));
                    y += 75;
                }
            }
            using (Pen pen = new Pen(Color.Black, 10))
            {
                pen.LineJoin = LineJoin.MiterClipped;
                pen.MiterLimit = 5.0f;
                DrawAngledLine(g, 20, pen, ptEnd);
                g.DrawString("Drag the end point to change join", 
                    this.Font, Brushes.Black, new Point(150, 100));
            }

        }

        private void DrawAngledLine(Graphics g, int y, Pen pen, Point pt)
        {

            g.DrawLines(pen, new Point[] { 
                    new Point (150, y),
                    new Point (250, y),
                    pt});
        }

        private void DrawJoinRect(Graphics g, Pen pen, Point pt)
        {
            RectangleF rect = new Rectangle(pt.X, pt.Y, 100, 50);
            g.DrawRectangle(pen, rect.X, rect.Y, rect.Width, rect.Height);
            //rect.Inflate((int)-pen.Width / 2, (int)-pen.Width / 2);
            StringFormat format = new StringFormat();
            format.Alignment = StringAlignment.Center;
            format.LineAlignment = StringAlignment.Center;
            g.DrawString(pen.LineJoin.ToString(), this.Font, Brushes.Black, rect, format);
        }

        private void MainFormCh5Pen_MouseDown(object sender, MouseEventArgs e)
        {

        }

        private void MainFormCh5Pen_MouseMove(object sender, MouseEventArgs e)
        {

        }

        private void MainFormCh5Pen_MouseUp(object sender, MouseEventArgs e)
        {

        }

        bool bDown = false;
        int posInflection = -1;
        Point ptEnd = new Point(180, 40);
        Point ptInflection = Point.Empty;
        Point ptMouseDown;
        GraphicsPath circle = new GraphicsPath();
        private void panelDraw_MouseDown(object sender, MouseEventArgs e)
        {
            if (DrawState.Join == drawState)
            {
                TestHit(ptEnd, e.Location);
            }
            else if (DrawState.DynamicBezier == drawState)
            {
                for (int i = 0; i < 4; i++)
                {
                    if (TestHit(ptsInflection[i], e.Location))
                    {
                        Console.WriteLine("pos inflection {0}; value {1}", i, ptsInflection[i]);
                        posInflection = i;
                        break;
                    }
                }
            }
        }

        private bool TestHit(Point pt, Point ptDown)
        {
            Rectangle rect = new Rectangle(
                pt - new Size(5, 5), new Size(10, 10));
            circle.Reset();
            circle.AddEllipse(rect);
            Console.WriteLine("ptDown {0}; rect {1}", ptDown, rect);
            if (circle.IsVisible(ptDown))
            {
                ptMouseDown = ptDown;
                bDown = true;
            }
            return bDown;
        }

        private void panelDraw_MouseMove(object sender, MouseEventArgs e)
        {
            if (drawState == DrawState.Join && bDown)
            {
                Point ptCurrentMouse = e.Location;
                ptEnd += new Size(ptCurrentMouse - new Size(ptMouseDown));
                ptMouseDown = e.Location;
                panelDraw.Invalidate();
            }
            if (drawState == DrawState.DynamicBezier && bDown)
            {
                if (posInflection != -1)
                {
                    Point ptCurrentMouse = e.Location;
                    ptsInflection[posInflection] += 
                        new Size(ptCurrentMouse - new Size(ptMouseDown));
                    ptMouseDown = e.Location;
                    panelDraw.Invalidate();
                }
            }
        }

        private void panelDraw_MouseUp(object sender, MouseEventArgs e)
        {
            if (drawState == DrawState.Join || drawState == DrawState.DynamicBezier)
            {
                bDown = false;
                posInflection = -1;
            }
        }


        private void DrawCurves(Graphics graphics)
        {
            using (Pen pen = new Pen(Color.Black, 4))
            {
                int width = 120;
                int height = 60;
                int margin = 15;

                ptInflection = new Point(margin, margin);

                graphics.DrawCurve(pen, new Point[] {
                    new Point(width - margin, margin), new Point(margin, margin),
                    new Point(width - margin, height - margin), new Point(margin, height - margin)},
                    0.0f);
                graphics.TranslateTransform(width, 0);
                graphics.DrawCurve(pen, new Point[] {
                    new Point(width - margin, margin), new Point(margin, margin),
                    new Point(width - margin, height - margin), new Point(margin, height - margin)},
                    0.5f);
                graphics.TranslateTransform(width, 0);
                graphics.DrawCurve(pen, new Point[] {
                    new Point(width - margin, margin), new Point(margin, margin),
                    new Point(width - margin, height - margin), new Point(margin, height - margin)},
                    1.0f);
                graphics.TranslateTransform(-2 * width, height);
                graphics.DrawString("Tension 0.0f", this.Font, Brushes.Black, margin, margin);
                graphics.TranslateTransform(width, 0);
                graphics.DrawString("Tension 0.5f", this.Font, Brushes.Black, margin, margin);
                graphics.TranslateTransform(width, 0);
                graphics.DrawString("Tension 1.0f", this.Font, Brushes.Black, margin, margin);                
                                
                graphics.TranslateTransform(-2 * width, height);
                Point[] pts = new Point[] {
                    new Point(margin, height-margin),
                    new Point(width-margin, height-margin),
                    new Point(margin, margin),
                    new Point(width-margin, margin)
                };                
                DrawLabeledBezier(graphics, pen, new Point[] { pts[0], pts[1], pts[2], pts[3] });
                graphics.TranslateTransform(width, 0);
                DrawLabeledBezier(graphics, pen, new Point[] { pts[2], pts[3], pts[1], pts[0] });
                graphics.TranslateTransform(width, 0);
                DrawLabeledBezier(graphics, pen, new Point[] { pts[2], pts[3], pts[0], pts[1] });

                graphics.SmoothingMode = SmoothingMode.None;
                graphics.TranslateTransform(-2 * width, height);
                graphics.DrawCurve(pen, new Point[] {
                    new Point(width - margin, margin), new Point(margin, margin),
                    new Point(width - margin, height - margin), new Point(margin, height - margin)},
                    0.5f);
                graphics.SmoothingMode = SmoothingMode.AntiAlias;
                graphics.TranslateTransform(width, 0);
                graphics.DrawCurve(pen, new Point[] {
                    new Point(width - margin, margin), new Point(margin, margin),
                    new Point(width - margin, height - margin), new Point(margin, height - margin)},
                    0.5f);                
            }
        }

        private void DrawLabeledBezier(Graphics g, Pen pen, Point[] pts)
        {
            pen.Color = Color.Black;
            pen.StartCap = LineCap.SquareAnchor;
            for (int i = 0; i < 4; i++)
            {
                g.DrawLine(pen, pts[i], pts[i] + new Size(1, 0));
                g.DrawString(i.ToString(), this.Font, Brushes.Black, pts[i]);
            }
            pen.StartCap = LineCap.NoAnchor;
            pen.Color = Color.Red;
            g.DrawBezier(pen, pts[0], pts[1], pts[2], pts[3]);            
        }
        
        static int m_nWidth = 120;
        static int m_nHeight = 60;
        static int m_nMargin = 15;
        Point[] ptsInflection = new Point[] {
                    new Point(m_nMargin, m_nHeight-m_nMargin),
                    new Point(m_nWidth-m_nMargin, m_nHeight-m_nMargin),
                    new Point(m_nMargin, m_nMargin),
                    new Point(m_nWidth-m_nMargin, m_nMargin)
                };
        private void DynamicBezier(Graphics graphics)
        {
            using (Pen pen = new Pen(Color.Black, 4))
            {
                DrawLabeledBezier(graphics, pen, ptsInflection);
            }
        }

        private void ClosedBezier(Graphics graphics)
        {
            using (GraphicsPath path = new GraphicsPath())
            using (Pen pen = new Pen(Color.Red, 2))
            {
                path.AddBezier(ptsInflection[0], ptsInflection[1], ptsInflection[2], ptsInflection[3]);
                path.CloseFigure();
                graphics.FillPath(Brushes.Gray, path);
                graphics.DrawPath(pen, path);
            }
        }

        private void panelDraw_Paint(object sender, PaintEventArgs e)
        {
            switch (drawState)
            {
                case DrawState.Compound:
                    DrawCompound(e.Graphics);
                    break;
                case DrawState.Join:
                    DrawJoin(e.Graphics);
                    break;
                case DrawState.Brushes:
                    DrawBrushes(e.Graphics);
                    break;
                case DrawState.DrawShapes:
                    DrawShapes(e.Graphics);
                    break;
                case DrawState.FillShapes:
                    FillShapes(e.Graphics);
                    break;
                case DrawState.DrawCurves:
                    DrawCurves(e.Graphics);
                    break;
                case DrawState.DynamicBezier:
                    DynamicBezier(e.Graphics);
                    break;
                case DrawState.ClosedBezier:
                    ClosedBezier(e.Graphics);
                    break;
            }
        }

        private void joinToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Text = "Join";
            drawState = DrawState.Join;
            panelDraw.Invalidate();
        }

        private void brushesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Text = "Pens from Brushes";
            drawState = DrawState.Brushes;
            Invalidate(true);
        }

        private void drawToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Text = "Draw Shapes";
            drawState = DrawState.DrawShapes;
            Invalidate(true);
        }

        private void fillToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Text = "Fill Shapes";
            drawState = DrawState.FillShapes;
            Invalidate(true);
        }

        private void curvesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Text = "Draw Curves";
            drawState = DrawState.DrawCurves;
            Invalidate(true);
        }

        private void dynamicBezierToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Text = "Dynamic Bezier";
            drawState = DrawState.DynamicBezier;
            Invalidate(true);
        }

        private void closeBezierToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Text = "Closed Bezier";
            drawState = DrawState.ClosedBezier;
            Invalidate(true);
        }
    }
}
